@@ -25,6 +25,7 @@
 #include <osmocom/core/byteswap.h>
 
 #include <osmocom/crypt/auth.h>
+#include <osmocom/crypt/kdf.h>
 
 #include <osmocom/gsm/tlv.h>
 #include <osmocom/gsm/gsm0808.h>
@@ -482,6 +483,7 @@ static int ran_a_decode_handover_request(struct ran_dec *ran_dec, const struct m
 	const struct tlv_p_entry *ie_aoip_transp_addr = TLVP_GET(tp, GSM0808_IE_AOIP_TRASP_ADDR);
 	const struct tlv_p_entry *ie_codec_list_msc_preferred = TLVP_GET(tp, GSM0808_IE_SPEECH_CODEC_LIST);
 	const struct tlv_p_entry *ie_call_id = TLVP_GET(tp, GSM0808_IE_CALL_ID);
+	const struct tlv_p_entry *ie_kc128 = TLVP_GET(tp, GSM0808_IE_KC_128);
 	const struct tlv_p_entry *ie_global_call_ref = TLVP_GET(tp, GSM0808_IE_GLOBAL_CALL_REF);
 
 	struct gsm0808_channel_type channel_type;
@@ -524,6 +526,11 @@ static int ran_a_decode_handover_request(struct ran_dec *ran_dec, const struct m
 			geran_encr.key_len = encr_info.key_len;
 		}
 
+		if (ie_kc128) {
+			memcpy(geran_encr.kc128, ie_kc128->val, 16);
+			geran_encr.kc128_present = true;
+		}
+
 		r->geran.chosen_encryption = &geran_encr;
 	}
 
@@ -997,7 +1004,8 @@ static struct msgb *ran_a_make_assignment_command(struct osmo_fsm_inst *log_fi,
 	if(ac->call_id_present == true)
 		call_id = &ac->call_id;
 
-	msg = gsm0808_create_ass(ac->channel_type, NULL, use_rtp_addr, use_scl, call_id);
+	msg = gsm0808_create_ass2(ac->channel_type, NULL, use_rtp_addr, use_scl, call_id,
+				  NULL, ac->lcls);
 	if (ac->osmux_present)
 		_gsm0808_assignment_extend_osmux(msg, ac->osmux_cid);
 	return msg;
@@ -1019,6 +1027,9 @@ static int a5_n_to_gsm0808_chosen_enc_alg(uint8_t *dst, int a5_n)
 	case 3:
 		*dst = GSM0808_ALG_ID_A5_3;
 		return 0;
+	case 4:
+		*dst = GSM0808_ALG_ID_A5_4;
+		return 0;
 	default:
 		return -ENOTSUP;
 	}
@@ -1058,14 +1069,17 @@ osmo_static_assert(sizeof(((struct gsm0808_encrypt_info*)0)->key) >= sizeof(((st
 		   gsm0808_encrypt_info_key_fits_osmo_auth_vec_kc);
 static struct msgb *ran_a_make_cipher_mode_command(struct osmo_fsm_inst *fi, const struct ran_cipher_mode_command *cm)
 {
-	struct gsm0808_encrypt_info ei = {};
+	struct gsm0808_cipher_mode_command cmc = {
+		.cipher_response_mode_present = true,
+		.cipher_response_mode = 1, /* 1: include IMEISV (3GPP TS 48.008 3.2.2.34) */
+	};
+	struct gsm0808_encrypt_info *ei = &cmc.ei;
 	char buf[16 * 2 + 1];
-	const uint8_t cipher_response_mode = 1;
 
-	if (make_encrypt_info_perm_algo(fi, &ei, cm->geran.a5_encryption_mask, cm->classmark))
+	if (make_encrypt_info_perm_algo(fi, ei, cm->geran.a5_encryption_mask, cm->classmark))
 		return NULL;
 
-	if (ei.perm_algo_len == 0) {
+	if (ei->perm_algo_len == 0) {
 		LOG_RAN_A_ENC(fi, LOGL_ERROR, "cannot start ciphering, no intersection between MSC-configured"
 			       " and MS-supported A5 algorithms. MSC: 0x%02x  MS: %s\n",
 			       cm->geran.a5_encryption_mask, osmo_gsm48_classmark_a5_name(cm->classmark));
@@ -1075,27 +1089,45 @@ static struct msgb *ran_a_make_cipher_mode_command(struct osmo_fsm_inst *fi, con
 	/* In case of UMTS AKA, the Kc for ciphering must be derived from the 3G auth
 	 * tokens.  vec->kc was calculated from the GSM algorithm and is not
 	 * necessarily a match for the UMTS AKA tokens. */
-	if (cm->geran.umts_aka)
-		osmo_auth_c3(ei.key, cm->vec->ck, cm->vec->ik);
-	else
-		memcpy(ei.key, cm->vec->kc, sizeof(cm->vec->kc));
-	ei.key_len = sizeof(cm->vec->kc);
+	if (cm->geran.umts_aka) {
+		int i;
+		osmo_auth_c3(ei->key, cm->vec->ck, cm->vec->ik);
+
+		for (i = 0; i < ei->perm_algo_len; i++) {
+			if (ei->perm_algo[i] != GSM0808_ALG_ID_A5_4)
+				continue;
+			/* A5/4 is included, so need to generate Kc128 */
+			osmo_kdf_kc128(cm->vec->ck, cm->vec->ik, cmc.kc128);
+			cmc.kc128_present = true;
+			break;
+		}
+	} else {
+		memcpy(ei->key, cm->vec->kc, sizeof(cm->vec->kc));
+	}
+	ei->key_len = sizeof(cm->vec->kc);
 
 	/* Store chosen GERAN key where the caller asked it to be stored.
 	 * alg_id remains unknown until we receive a Cipher Mode Complete from the BSC */
 	if (cm->geran.chosen_key) {
-		if (ei.key_len > sizeof(cm->geran.chosen_key->key)) {
+		*cm->geran.chosen_key = (struct geran_encr){0};
+
+		if (ei->key_len > sizeof(cm->geran.chosen_key->key)) {
 			LOG_RAN_A_ENC(fi, LOGL_ERROR, "Chosen key is larger than I can store\n");
 			return NULL;
 		}
-		memcpy(cm->geran.chosen_key->key, ei.key, ei.key_len);
-		cm->geran.chosen_key->key_len = ei.key_len;
+		memcpy(cm->geran.chosen_key->key, ei->key, ei->key_len);
+		cm->geran.chosen_key->key_len = ei->key_len;
+
+		if (cmc.kc128_present) {
+			memcpy(cm->geran.chosen_key->kc128, cmc.kc128, 16);
+			cm->geran.chosen_key->kc128_present = true;
+		}
 	}
 
 	LOG_RAN_A_ENC(fi, LOGL_DEBUG, "Tx BSSMAP CIPHER MODE COMMAND to BSC, %u ciphers (%s) key %s\n",
-		       ei.perm_algo_len, osmo_hexdump_nospc(ei.perm_algo, ei.perm_algo_len),
-		       osmo_hexdump_buf(buf, sizeof(buf), ei.key, ei.key_len, NULL, false));
-	return gsm0808_create_cipher(&ei, cm->geran.retrieve_imeisv ? &cipher_response_mode : NULL);
+		       ei->perm_algo_len, osmo_hexdump_nospc(ei->perm_algo, ei->perm_algo_len),
+		       osmo_hexdump_buf(buf, sizeof(buf), ei->key, ei->key_len, NULL, false));
+	return gsm0808_create_cipher2(&cmc);
 }
 
 struct msgb *ran_a_make_handover_request(struct osmo_fsm_inst *log_fi, const struct ran_handover_request *n)
@@ -1141,6 +1173,12 @@ struct msgb *ran_a_make_handover_request(struct osmo_fsm_inst *log_fi, const str
 		       n->geran.chosen_encryption->key, n->geran.chosen_encryption->key_len);
 		r.encryption_information.key_len = n->geran.chosen_encryption->key_len;
 		r.chosen_encryption_algorithm_serving = n->geran.chosen_encryption->alg_id;
+
+		if (n->geran.chosen_encryption->kc128_present) {
+			r.more_items = true;
+			memcpy(r.kc128, n->geran.chosen_encryption->kc128, sizeof(r.kc128));
+			r.kc128_present = true;
+		}
 	}
 
 	if (n->classmark)
@@ -1218,7 +1256,11 @@ static struct msgb *_ran_a_encode(struct osmo_fsm_inst *caller_fi, const struct
 		return ran_a_make_assignment_command(caller_fi, &ran_enc_msg->assignment_command);
 
 	case RAN_MSG_COMMON_ID:
-		return gsm0808_create_common_id(ran_enc_msg->common_id.imsi, NULL, NULL);
+		return gsm0808_create_common_id(ran_enc_msg->common_id.imsi, NULL,
+						ran_enc_msg->common_id.last_eutran_plmn_present ?
+							&ran_enc_msg->common_id.last_eutran_plmn :
+							NULL
+						);
 
 	case RAN_MSG_CIPHER_MODE_COMMAND:
 		return ran_a_make_cipher_mode_command(caller_fi, &ran_enc_msg->cipher_mode_command);
