@@ -127,11 +127,9 @@ static void print_help()
 	printf("  -D --daemonize             Fork the process into a background daemon.\n");
 	printf("  -c --config-file filename  The config file to use.\n");
 	printf("  -s --disable-color\n");
-	printf("  -l --database db-name      The database to use.\n");
 	printf("  -T --timestamp             Prefix every log line with a timestamp.\n");
 	printf("  -V --version               Print the version of OsmoMSC.\n");
 	printf("  -e --log-level number      Set a global loglevel.\n");
-	printf("  -M --mncc-sock-path PATH   Disable built-in MNCC handler and offer socket.\n");
 
 	printf("\nVTY reference generation:\n");
 	printf("     --vty-ref-mode MODE     VTY reference generation mode (e.g. 'expert').\n");
@@ -174,11 +172,11 @@ static void handle_options(int argc, char **argv)
 			{"daemonize", 0, 0, 'D'},
 			{"config-file", 1, 0, 'c'},
 			{"disable-color", 0, 0, 's'},
-			{"database", 1, 0, 'l'},
+			{"database", 1, 0, 'l'}, /* deprecated */
 			{"timestamp", 0, 0, 'T'},
 			{"version", 0, 0, 'V' },
 			{"log-level", 1, 0, 'e'},
-			{"mncc-sock-path", 1, 0, 'M'},
+			{"mncc-sock-path", 1, 0, 'M'}, /* deprecated */
 			{"no-dbcounter", 0, 0, 'C'}, /* deprecated */
 			{"vty-ref-mode", 1, &long_option, 1},
 			{"vty-ref-xml", 0, &long_option, 2},
@@ -209,6 +207,8 @@ static void handle_options(int argc, char **argv)
 			break;
 		case 'l':
 			msc_cmdline_config.database_name = optarg;
+			fprintf(stderr, "Command line argument '-%c' is deprecated, use VTY "
+				"parameter 'msc' / 'sms-database %s' instead.\n", c, optarg);
 			break;
 		case 'c':
 			msc_cmdline_config.config_file = optarg;
@@ -221,6 +221,8 @@ static void handle_options(int argc, char **argv)
 			break;
 		case 'M':
 			msc_cmdline_config.mncc_sock_path = optarg;
+			fprintf(stderr, "Command line argument '-%c' is deprecated, use VTY "
+				"parameter 'msc' / 'mncc external %s' instead.\n", c, optarg);
 			break;
 		case 'C':
 			fprintf(stderr, "-C is deprecated and does nothing.");
@@ -258,6 +260,7 @@ struct gsm_network *msc_network_alloc(void *ctx,
 
 	mgcp_client_conf_init(&net->mgw.conf);
 	net->call_waiting = true;
+	net->lcls_permitted = false;
 
 	net->mgw.tdefs = g_mgw_tdefs;
 	osmo_tdefs_reset(net->mgw.tdefs);
@@ -776,14 +779,24 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 		}
 	}
 
-	while (!quit) {
+	do {
 		log_reset_context();
 		osmo_select_main_ctx(0);
-	}
 
-	msc_network_shutdown(msc_network);
-	osmo_signal_dispatch(SS_L_GLOBAL, S_L_GLOBAL_SHUTDOWN, NULL);
-	sleep(3);
+		/* If the user hits Ctrl-C the third time, just terminate immediately. */
+		if (quit >= 3)
+			break;
+
+		/* Has SIGTERM been received (and not yet been handled)? */
+		if (quit && !osmo_select_shutdown_requested()) {
+			msc_network_shutdown(msc_network);
+			osmo_signal_dispatch(SS_L_GLOBAL, S_L_GLOBAL_SHUTDOWN, NULL);
+
+			/* Request write-only mode in osmo_select_main_ctx() */
+			osmo_select_shutdown_request();
+			/* continue the main select loop until all write queues are serviced. */
+		}
+	} while (!osmo_select_shutdown_done());
 
 	log_fini();
 
